/***************************************************************************
 *   Copyright (C) 2006 - 2009 by Yoshihiro Ota
 *   ota@j.email.ne.jp
 *                                                                         
 *   This program is free software; you can redistribute it and/or modify  
 *   it under the terms of the GNU General Public License as published by  
 *   the Free Software Foundation; either version 2 of the License, or     
 *   (at your option) any later version.                                   
 *                                                                         
 *   This program is distributed in the hope that it will be useful,       
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
 *   GNU General Public License for more details.                          
 *                                                                         
 *   You should have received a copy of the GNU General Public License     
 *   along with this program; if not, write to the                         
 *   Free Software Foundation, Inc.,                                       
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             
 *
 ***************************************************************************
 *
 * Algorithm:
 *  Searching for files with the same contents is 3 stage process.
 *  Gather All File Information
 *  1. Traverse all files and directories given to the command line option.
 *  1.1 Obtain file information such as i-node, path, and timestamps.
 *  1.2 Grope files by a pair of device id and file size.  Unless files are
 *      in the same grope, the content of files won't match.
 *  Match same content candidates
 *  2. Find files with same contents and link them together.
 *  2.1 First, sort a group by i-node.
 *  2.2 Check which files contains the same context.
 *  2.2.1 Check if file status allows to combine files together.
 *  2.2.1.1 Check if owner informations are the same.
 *  2.2.1.2 Check if permissions are the same. // not done yet
 *  2.3 Read files by i-node and check if contents are the same.
 *      Files are only read upon necessity.
 *  2.3.1 Check first 512 bytes if they are the same.
 *  2.3.2 Calculate checksums and compare. 
 *  2.3.2.1 MD5 is required.
 *  2.3.2.2 SHAs are available in FreeBSD. // need to be configurable
 *  2.3.3 Unlink and copy.
 *  2.3.4 Unlink and link.
 *         Note that there may be more than one newer files but all newer
 *         files have the same i-node.  Since files are read by i-node but
 *         not filename, the same i-node file is read at most one.
 *  2.4 Cleanup memory allocated for a group.
 *  3 Restore timestamps.
 *
 ***************************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h> /* vfprintf */
#include <string.h>
#include <errno.h> /* errno */
#include <time.h> /* difftime */
#include <math.h> /* ceil */

#include <sys/param.h> /* MAXPATHLEN */

#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h> /* link/unlink */ /* *dir */

#include <utime.h> /* utime */
#include <dirent.h> /* dirent */

#include <signal.h> /* signal */
#include "binsert.h"

typedef enum { false, true } boolean;

typedef struct
{
	int num; /* number of entries usd. */
	int max; /* maximum */
	void **entries; /* array of pointers */
} dynamic_array_t;

dynamic_array_t* init_array(dynamic_array_t *array, int size)
{
	array->num = 0;
	array->max = size;
	array->entries = malloc(array->max * sizeof(*array->entries));
	if(array->entries == NULL)
		return NULL;
	return array;
}

dynamic_array_t* resize_array(dynamic_array_t *array, int new_size)
{
	void *ptr = realloc(array->entries,
				new_size * sizeof(*array->entries));
	if(ptr == NULL)
	{
		perror("resize_array:realloc:out of memory");
		exit(-__LINE__);
	}
	array->entries = ptr;
	array->max = new_size;
	return array;
}

dynamic_array_t* reserve_index(dynamic_array_t *array, int index)
{
	if(array->max < index)
	{
		int new_size = 2 * array->max;
		return resize_array(array,
			((index > new_size) ? index + array->max : new_size));
	}
	return array;
}

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

// #define __DEBUG__
#if defined(__DEBUG__)
#define aprintf(txt, ...) printf("%s:" txt, __func__, __VA_ARGS__)
#define X(x)	fprintf(stderr, #x " %d\n", x)
#else
#define aprintf(txt, ...)
#define X(x)
#endif

#include <md5.h>
#ifdef HAVE_SHA
#include <sha.h>
#include <sha256.h>
#endif

#if LINK_MAX <= 0
#error LINK_MAX not defined
#endif
int link_max = LINK_MAX / 4; /* reserve some extra space for others */
char *inputfile = NULL;
char *outputfile = NULL;

typedef enum
{
	shell_i = 's', /* shell script */
	progress_i = 'P', /* stage progress */
	list_i = 'L', /* list files at each stage */
	stats_i = 'S', /* statistics at end */
	action_i = 'A', /* unlink, link, and cp */
	domain_i = 'D', /* domain related */

	files_i = 'f', /* file processing */
	dirs_i = 'd', /* directory processing */
} logging_x;

typedef struct path_x
{
	char *name;
	boolean is_dir:1;
	boolean is_link_src:1;
	struct utimbuf time;
	struct path_x *next; /* stack: first-in, last-out */
} path_x;

typedef struct inode_x
{
	ino_t st_ino;
	off_t st_size;
	nlink_t st_nlink;

	char head[512];
	int head_len;

	boolean sig_ready; /* signature has been genereated */
	char tail[512];
	int tail_len;
#ifdef PRINT
	char md5sum[33];
  #ifdef HAVE_SHA
	char sha1sum[41];
	char sha256sum[65];
  #endif
#else
	unsigned char md5sum[16];
  #ifdef HAVE_SHA
	unsigned char sha1sum[20];
	unsigned char sha256sum[32];
  #endif
#endif
	struct inode_x *next;
} inode_x;

typedef struct file_x
{
	path_x *path;
	struct stat stat;
	boolean stat_checked;
	inode_x *inode;
	struct file_x *file; /* when same inode is shared, they create a list */
} file_x;

typedef struct
{
	dev_t st_dev; /* device ID */
	off_t st_size; /* file size */

	dynamic_array_t files;
	inode_x *inode; /* linkted-list stack */
} domain_x; /* device data */


typedef struct
{
	unsigned int use_cmp:1;
	unsigned int execute:1;
	unsigned int restore_timestamp:1;

	char *debug;
	dynamic_array_t domain;
	path_x *access_time; /* stack of files and dirs : it maintains a list
		of directories and files to restore timestamps */
	struct
	{
		uint64_t files;
		uint64_t dirs;
		uint64_t linked;
	} stat;
} system_x;
system_x sys = { 0, 1, 0, NULL,
	{ 0, 512, NULL}, NULL, {0, }, };

int compare_files_by_inode(const void *data1, const void *data2)
{ /* desending order by link-count and then asending order by inode */
	const file_x * file1 = *(file_x * const*) data1;
	const file_x * file2 = *(file_x * const*) data2;

	if(file2->stat.st_nlink != file1->stat.st_nlink)
		return file2->stat.st_nlink - file1->stat.st_nlink;
	return file1->stat.st_ino - file2->stat.st_ino;
}

void info(logging_x level, const char *format, ...)
{
	if(sys.debug != NULL && strchr(sys.debug, level))
	{
		va_list args;
		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
	}
}

domain_x* new_domain(dev_t st_dev, off_t st_size)
{
	domain_x *domain;
	domain = calloc(1, sizeof(*domain));
	if(domain == NULL)
	{
		perror("new_domain:calloc:out of memory");
		exit(1);
	}
	domain->st_dev = st_dev;
	domain->st_size = st_size;
	if(init_array(&domain->files, 16) == NULL)
	{
		char buf[512];
		sprintf(buf, "%s:realloc:out of memory", __func__);
		perror(buf);
		exit(-__LINE__);
	}
	info(domain_i, "new domain(%hu, %lld)\n", st_dev, st_size);
	return domain;
}

int compare_domains(const void *data1, const void *data2)
{
	const domain_x * domain1 = *(domain_x * const*) data1;
	const domain_x * domain2 = *(domain_x * const*) data2;
	if(domain1->st_dev < domain2->st_dev)
		return 1;
	else if(domain1->st_dev < domain2->st_dev)
		return -1;
	else
	{
		if(domain1->st_size < domain2->st_size)
			return 1;
		else if(domain1->st_size > domain2->st_size)
			return -1;
		else
			return 0;
	}
}

domain_x* find_domain(file_x *file)
{
	int i;
	domain_x data, *key = &data, **found;
	key->st_dev = file->stat.st_dev;
	key->st_size = file->stat.st_size;
	found = bsearch(&key, sys.domain.entries, sys.domain.num,
		sizeof(sys.domain.entries[0]), compare_domains);
	if(found != NULL)
		return *found;

	reserve_index(&sys.domain, sys.domain.num + 1);
	key = sys.domain.entries[sys.domain.num++] =
		new_domain(file->stat.st_dev, file->stat.st_size);
	binsert(sys.domain.entries, sys.domain.num,
		sizeof(sys.domain.entries[0]), compare_domains);
	return key;
}

inode_x* find_inode(domain_x *domain, ino_t st_ino)
{
	inode_x *inode = domain->inode;
	while(inode != NULL)
		if(st_ino == inode->st_ino)
			return inode;
		else
			inode = inode->next;
	return NULL;
}

int add_new_path(path_x *access_time)
{
	access_time->next = sys.access_time;
	sys.access_time = access_time;
	return 1;
}

void restore_timestamps(int sig)
{
	path_x *path;
	if(sig)
	{
		fprintf(stderr, "signal %d\n", sig);
		fprintf(stderr, "restoring timestamps...\n", sig);
	}

	while(path = sys.access_time)
	{
		if(path->is_dir || /* restore directory timestamps */
			path->is_link_src) /* restore file timestamps */
		{
			if(sys.restore_timestamp)
				utime(path->name, &path->time);
		}
#if defined(FREE)
		if(path->name)
			free(path->name);
#endif
		sys.access_time = path->next;
#if defined(FREE)
		free(path);
#endif
	}
#if defined(FREE)
	free(sys.domain.entries);
#endif
	exit(sig);
}

file_x* create_filex(path_x *path, struct stat *stat) 
{
	file_x *file;

	if(!(file = calloc(1, sizeof(*file))))
	{
		perror("create_filex:calloc");
		return NULL;
		// return errno;
	}

	file->path = path;
	file->stat = *stat;
	return file;
}

int add_new_file(file_x *file)
{
	int i;
	domain_x *domain;
	domain = find_domain(file);

	reserve_index(&domain->files, domain->files.num + 1);
	domain->files.entries[domain->files.num++] = file;
#if 0
/* use qsort when needed */
	binsert(domain->files, domain->num,
		sizeof(domain->files[0]), compare_files_by_inode);
#endif
	return 1;
}

int free_file(file_x **file)
{
#if defined(FREE)
	free((*file)->path->name);
	(*file)->path->name = NULL;
	free(*file);
	*file = NULL;
#endif
}

int free_files(file_x **file)
{
	file_x *next;
	do
	{
		next = (*file)->file;
		free_file(file);
		file = &next;
	}while(file);
}

int file_stat_allowed(file_x *one, file_x *another)
{
	if(one->stat.st_dev != another->stat.st_dev)
		return 0; /* false */
	if(one->stat.st_uid != another->stat.st_uid)
		return 0; /* false */
	if(one->stat.st_gid != another->stat.st_gid)
		return 0; /* false */
	return 1; /* true */
}

int add_inodex(domain_x *domain, file_x *file)
{
	if(!(file->inode = calloc(1, sizeof(*(file->inode)))))
	{
		char buf[512];
		sprintf(buf, "%s:calloc()", __func__);
		perror(buf);
		return 0; /* false */
	}
	file->inode->st_ino = file->stat.st_ino;
	file->inode->st_size = file->stat.st_size;
	file->inode->st_nlink = file->stat.st_nlink; /* this is the first time
		to refrence this file with this i-node. thus, it is safe
		to assume the number of hard links hasn't changed by this
		program. This is not a good assumption if other programs
		are running... */
	file->inode->next = domain->inode;
	domain->inode = file->inode;
	return 1; /* true */
}

int read_head(domain_x *domain, file_x *file)
{
	FILE *fd;
	if(file->inode == NULL)
	{
		file->inode = find_inode(domain, file->stat.st_ino);
		if(file->inode == NULL)
			if(!add_inodex(domain, file)) /* XXX: error handling */
				return 0; /* false */
	}
	if(file->stat.st_size == 0) /* no contents to read */
		return 1; /* true */
	if(file->inode->head_len)
		return 1; /* true */
	fd = fopen(file->path->name, "r");
	if(fd == NULL)
	{
		char buf[512];
		sprintf(buf, "%s:fopen(%s)", __func__, file->path->name);
		perror(buf);
		return 0; /* false */
	}
	file->inode->head_len =
		fread(file->inode->head, 1, sizeof(file->inode->head), fd);
	fclose(fd);
	return file->inode->head_len; /* true */
}

int same_head(domain_x *domain, file_x *one, file_x *another)
{
	if(!read_head(domain, one))
		return 0;
	if(!read_head(domain, another))
		return 0;
	if(one->inode->head_len != another->inode->head_len)
		return 0;
	return !memcmp(one->inode->head, another->inode->head,
		one->inode->head_len);
}

int read_sig(file_x *file)
{
	FILE *fd;
	MD5_CTX md5;
#ifdef HAVE_SHA
	SHA1_CTX sha1;
	SHA256_CTX sha256;
#endif
	int len = 0;
	if(file->inode->sig_ready)
		return 1; /* true */
	/* inode->head contains all contents; no need to compare signatures */
	if(file->inode->st_size <= sizeof(file->inode->head))
		return 1; /* true */
	fd = fopen(file->path->name, "r");
	if(fd == NULL)
	{
		char buf[512];
		sprintf(buf, "%s:fopen(%s)", __func__, file->path->name);
		perror(buf);
		return 0; /* false */
	}
	MD5Init(&md5);
#ifdef HAVE_SHA
	SHA1_Init(&sha1);
	SHA256_Init(&sha256);
#endif
	do
	{
		file->inode->tail_len = len;
		len = fread(file->inode->tail, 1, sizeof(file->inode->tail), fd);
		MD5Update(&md5, file->inode->tail, len);
#ifdef HAVE_SHA
		SHA1_Update(&sha1, file->inode->tail, len);
		SHA256_Update(&sha256, file->inode->tail, len);
#endif
	}while(!feof(fd));
	if(len != 0)
		file->inode->tail_len = len;
#ifdef PRINT
	MD5End(&md5, file->inode->md5sum);
#ifdef HAVE_SHA
	SHA1_End(&sha1, file->inode->sha1sum);
	SHA256_End(&sha256, file->inode->sha256sum);
#endif
#else
	MD5Final(file->inode->md5sum, &md5);
#ifdef HAVE_SHA
	SHA1_Final(file->inode->sha1sum, &sha1);
	SHA256_Final(file->inode->sha256sum, &sha256);
#endif
#endif
	fclose(fd);
#if 0
	aprintf("path(%s) -> md5(%s)\n", file->path, file->inode->md5sum);
#endif
	file->inode->sig_ready = true;
	return 1; /* true */
}

int same_contents(file_x *one, file_x *another)
{
	/* inode->head contains all contents; no need to compare signatures */
	/* same_head() must be called prior to same_contents() */
	if(one->inode->st_size <= sizeof(one->inode->head))
		return 1; /* true */
	read_sig(one);
	read_sig(another);
	if(one->inode->tail_len != another->inode->tail_len)
		return 0; /* false */
	if(memcmp(one->inode->tail, another->inode->tail,
		one->inode->tail_len) != 0)
		return 0; /* false */
	if(memcmp(one->inode->md5sum, another->inode->md5sum,
		sizeof(one->inode->md5sum)) != 0)
		return 0; /* false */
#ifdef HAVE_SHA
	if(memcmp(one->inode->sha1sum, another->inode->sha1sum,
		sizeof(one->inode->sha1sum)) != 0)
		return 0; /* false */
	if(memcmp(one->inode->sha256sum, another->inode->sha256sum,
		sizeof(one->inode->sha256sum)) != 0)
		return 0; /* false */
#endif
	return 1; /* true */
}

#define NOT_SAME 0
#define SAME_INO 1
#define SAME_CONTENTS 2
int same_file(domain_x *domain, file_x *one, file_x *another)
{
	if(one == NULL || another == NULL)
		return NOT_SAME; /* file was already processed; false */

	if(one->stat.st_ino == another->stat.st_ino) /* ignore same inode */
	{
		return SAME_INO; /* changing the behavior 2008/03/17 */
		// free_file(&domain->files[another]);
		return NOT_SAME; /* it is same file but already has same inode */
	}

	if(!file_stat_allowed(one, another))
		return NOT_SAME; /* false */
	if(!same_head(domain, one, another))
		return NOT_SAME; /* false */
	if(!same_contents(one, another))
		return NOT_SAME; /* false */
	return SAME_CONTENTS;
}

int link_file(domain_x *domain, file_x *one, file_x *another)
{
	if(one->inode->st_nlink > link_max)
	{
		info(action_i, "link_max exceesed (%s[%u] %d)\n",
			one->path->name,
			one->stat.st_ino,
			one->inode->st_nlink);
		/* free_file(&domain->files[one]); /* check this later */
	}
	else
	{
		info(action_i, "unlink(%s[%u]); link(%s[%u], %s[%u])\n",
			another->path->name,
			another->stat.st_ino,
			one->path->name,
			one->stat.st_ino,
			another->path->name,
			another->stat.st_ino);
		info(shell_i, "rm %s; ln %s %s\n",
			another->path->name,
			one->path->name,
			another->path->name);
		if(sys.execute)
			if(unlink(another->path->name))
			{
				char buf[512];
				sprintf(buf, "%s:unlink(%s)", __func__, another->path->name);
				perror(buf);
			exit(999);
			}
		another->inode->st_nlink--;
		one->path->is_link_src = 1;
		if(sys.execute)
			if(link(one->path->name, another->path->name))
			{
				char buf[512];
				sprintf(buf, "%s:link(%s, %s)", __func__,
					one->path->name, another->path->name);
				perror(buf);
			}
		one->inode->st_nlink++;
		sys.stat.linked++;
		// if(sys.execute)
			// free_file(&domain->files[another]);
	}
}

int clone_file(domain_x *domain, file_x *one, file_x *another)
{
	char buffer[3 * MAXPATHLEN];

	if(sys.execute)
		if(unlink(another->path->name))
		{
			char buf[512];
			sprintf(buf, "%s:unlink(%s)", __func__, another->path->name);
			perror(buf);
			exit(999);
		}
	another->inode->st_nlink--;
	snprintf(buffer, sizeof(buffer), "cp -p %s %s",
		one->path->name, another->path->name);
	info(action_i, "%s\n", buffer);
	info(shell_i, "%s\n", buffer);
	if(sys.execute)
		if(system(buffer))
		{
			char buf[512];
			sprintf(buf, "%s:systen(%s)", __func__, buffer);
			perror(buf);
			return -__LINE__; /* false */
		}
	if(lstat(another->path->name, &another->stat) == -1)
	{ /* XXX: this does not work with dry-run */
		char buf[512];
		sprintf(buf, "%s:lstat(%s)", __func__, another->path->name);
		perror(buf);
		return -__LINE__;
	}

	another->file = NULL; /* new standalone file */
	if(!add_inodex(domain, another))
		return -__LINE__;
	return -__LINE__;
}

void free_domain(domain_x *domain)
{
	int i;
	inode_x *inode;

#if defined(FREE)
/* FREE is memory actually hart the peformance if large portion on the
process in the swap space... So, don't free it, yet.  */

	info(domain_i, "free domain(%hu, %lld)\n",
		domain->st_dev, domain->st_size);
	for(i = 0; i < domain->files.num; i++)
		if(domain->files.entries[i] != NULL)
		{
#if 0
/* what is this for? */
			if(sys.restore_timestamp)
				utime(domain->files[i]->path->name,
					&domain->files[i]->path->time);
			if(domain->files[i]->path->name)
			{
				free(domain->files[i]->path->name);
				domain->files[i]->path->name = NULL;
			}
#endif
			free_file(&domain->files.entries[i]);
		}
#endif
	while(domain->inode) /* we want to free inodes */
	{
		inode = domain->inode;
		domain->inode = domain->inode->next;
/**/
{
	static FILE *fd = NULL;
	size_t size;
	if(fd == NULL && outputfile != NULL)
		fd = fopen(outputfile, "w");
	if(fd != NULL)
	{
		size = fwrite(inode, sizeof(*inode), 1, fd);
		// printf("size = %d\r", size);
	}
}
/**/
		free(inode);
	}
#if defined(FREE)
	free(domain);
#endif
}

int inoctl_by_domain(domain_x *domain) /* return value is currently not used */
{
	int i = 0, j, k;
	int count = 0; /* the number of unique i-nodes */
	int ino;
	file_x **files = (file_x**) domain->files.entries;

	/* count will be the minimum; domain->files.num will be the maximum */
	file_x *heads[domain->files.num]; /* array of pointers to pointer */
	inode_x *inodes[domain->files.num]; /* array of pointers to pointer */
	int same[domain->files.num];

	qsort(domain->files.entries, domain->files.num,
		sizeof(domain->files.entries[0]), compare_files_by_inode);

	/* create a list for each inode. */
	do /* i for domain->files[] */
	{
		if(!add_inodex(domain, files[i]))
			exit(-__LINE__);
		heads[count] = files[i];
		inodes[count] = files[i]->inode;
		for(i++;
			i < domain->files.num &&
				files[i - 1]->stat.st_ino == files[i]->stat.st_ino;
			i++)
		{
			files[i]->inode = inodes[count];
			files[i - 1]->file = files[i];
		}
		count++;
	}while(i < domain->files.num);

	for(i = 0; i < count; i++)
	{
		int dup = 0, total = 0;

		if(heads[i] == NULL) /* skip ones deleted */
			continue;

		memset(same, 0, sizeof(same));
		total += inodes[i]->st_nlink;
		same[dup++] = i;
		for(j = i + 1; j < count; j++)
		{
			if(heads[j] == NULL) /* skip ones deleted */
				continue;
			switch(same_file(domain, heads[i], heads[j]))
			{
				case NOT_SAME:
					continue; /* go back to the loop */
				case SAME_CONTENTS:
					total += inodes[j]->st_nlink;
					same[dup++] = j;
					if(sys.use_cmp)
					{
						int status;
						char buffer[2 * MAXPATHLEN + 64];
						snprintf(buffer, sizeof(buffer), "cmp %s %s",
						heads[i]->path->name, heads[j]->path->name,
						heads[i]->path->name, heads[j]->path->name);
						status = system(buffer);
						if(status)
							printf("%d \t%s \t%s\n", status,
								inodes[i]->head, inodes[j]->head);
					}
					break;
			}
		}

	if(sys.debug != NULL && strchr(sys.debug, list_i))
	{
		info(list_i, "total inodes = %d, total files = %d link_max = %d\n",
			dup, total, link_max);
		for(j = 0; j < dup; j++)
		{
			inode_x *inode = inodes[same[j]];
			file_x *file;
			info(list_i, " inode ino = %u nlink = %d\n",
				inode->st_ino, inode->st_nlink);
			for(file = heads[same[j]]; file; file = file->file)
				info(list_i, "  %s\n", file->path->name);
		}
	}
#if 1
		for(k = 0; k < dup; k++) /* first inode which has less than link_max */
			if(inodes[same[k]]->st_nlink < link_max)
				break;

		for(j = 0; j < k; j++)
		{
			if(k == dup)
				goto clone;
			while(inodes[same[j]]->st_nlink > link_max)
			{
				if(inodes[same[k]]->st_nlink >= link_max)
				{
					k++;
					goto clone;
				}
				link_file(domain, heads[same[k]], heads[same[j]]); /* check error */
				heads[same[j]] = heads[same[j]]->file;
				if(heads[same[k]]->inode->st_nlink >= link_max)
				{
					k++;
					clone:
					if(dup <= k)
					{
						file_x *file = heads[same[j]];
						heads[same[j]] = heads[same[j]]->file;
						clone_file(domain, heads[same[j]], file); /* check error */
						same[k] = k;
						heads[k] = file;
						inodes[k] = file->inode;
						dup++; // Is this correct?  Do I take care of this right?
						// count++;
					}
				}
			}
		}
#endif
		for(k = dup - 1; j < k; j++) /* 0, j, 2, --> ... <--, k, 7, 8 */
		{
			while(inodes[same[j]]->st_nlink < link_max)
			{
				link_file(domain, heads[same[j]], heads[same[k]]); /* check error */
				heads[same[k]] = heads[same[k]]->file;
				while(heads[same[k]] == NULL)
				{
					k--;
					if(j >= k)
						goto done;
				}
			}
		}
		done:
		for(j = 0; j < dup; j++)
			heads[same[j]] = NULL;
	}

	free_domain(domain);
	return i;
}

int inoctl()
{
	int i = 0;
	info(progress_i, "1st stage completed: gathered all filenames\n");
	for(i = 0; i < sys.domain.num; i++)
	{
		domain_x *domain = sys.domain.entries[i];
		file_x *file = domain->files.entries[0];
		info(domain_i,
			"process domain(%hu, %lld) %d files, %d/%d domains\n",
			domain->st_dev, domain->st_size, domain->files.num,
			i, sys.domain.num);
		inoctl_by_domain(domain);
	}
	info(progress_i, "2nd stage completed: combined same files\n");
	restore_timestamps(0);
	info(progress_i, "3rd stage completed: restored timestamps\n");
	info(stats_i,
		"%llu files\n"
		"%llu dirs\n"
		"%llu linked\n"
		,
		sys.stat.files,
		sys.stat.dirs,
		sys.stat.linked,
		0);
}

int loopdir(char *path,
	int (*callback)(char *parent, struct dirent *entry, void *data),
	void *data)
{
	DIR *dir;
    struct dirent *entry;

	dir = opendir(path);
	if(dir != NULL)
	{
		while((entry = readdir(dir)) != NULL)
		{
			if(entry->d_name[0] == '.')
			{
				if(entry->d_name[1] == 0)
					continue;
				else if(entry->d_name[1] == '.' &&
					entry->d_name[2] == 0)
					continue;
			}
			callback(path, entry, data);
		}
		closedir(dir);
	}
}

/*  Retreve files and directoires to store meta-data. */
int retreave(char *parent, struct dirent *entry, void *data)
{
	char path[MAXPATHLEN + 1];
	path_x *orig_time;
	struct stat stat;

	orig_time = calloc(1, sizeof(*orig_time));
	if(entry)
		snprintf(path, sizeof(path), "%s/%s", parent, entry->d_name);
	else
		snprintf(path, sizeof(path), "%s", parent);

	if(lstat(path, &stat) == -1)
	{
		char buf[512];
		sprintf(buf, "%s:lstat(%s)", __func__, path);
		perror(buf);
		return -__LINE__;
	}
	orig_time->name = strdup(path); /* XXX: check if NULL */
	orig_time->time.actime = stat.st_atime;
	orig_time->time.modtime = stat.st_mtime;
	if(S_ISREG(stat.st_mode))
	{
		file_x *file;

		info(files_i, "file %s\n", orig_time->name);
		file = create_filex(orig_time, &stat);
		if(file == NULL)
			return errno; /* XXX: bad fasion of returning error */

		add_new_path(orig_time);
		add_new_file(file);
		sys.stat.files++;
	}
	else if(S_ISDIR(stat.st_mode))
	{
		orig_time->is_dir = true;
		add_new_path(orig_time);
		info(dirs_i, "dir %s\n", orig_time->name);
		loopdir(orig_time->name, retreave, NULL);
		sys.stat.dirs++;
	}
	else /* XXX: revisit behavoir with sym-links. */
	{
		free(orig_time->name);
		free(orig_time);
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int i;
	char c;
	char *filename = NULL;

	if(argc < 2)
		goto help; /* XXX: temoprary */

	/* -i and -o are not yet supported */
	while((c = getopt(argc, argv, "cd:f:hm:nt" "i:o:")) != -1)
	{
		switch(c)
		{
			case 'c':
				sys.use_cmp = 1;
				break;
			case 'd':
				sys.debug = optarg;
				break;
			case 'f':
				filename = optarg;
				break;
			case 'm':
				link_max = atoi(optarg);
				break;
			case 'n':
				sys.execute = 0;
				break;
			case 't':
				sys.restore_timestamp = 1;
				break;
			case 'h':
			help:
	fprintf(stderr,
		"ffunsion [<directories | files>]\n"
		"-c : use 'cmp' to verify the contents of files (expensive)\n"
		"-d : debug\n"
			"\t's' shell script (broken...)\n"
			"\t'P' stage Progress\n"
			"\t'L' List files at each stage\n"
			"\t'S' Statistics at end\n"
			"\t'A' Actions such as unlink, link, cp\n"
			"\t'D' Domain related\n"
			"\t'f' File processing\n"
			"\t'd' Directory processing\n"
		"-f : input from a file or stdin (only one file can be specified,\n"
			"\tfiles/directories from argument will be ignored.)\n"
		"-m : link_max\n"
		"-n : no-execute; dry-run\n"
		"-t : restore timestamps\n"
		"-h : help\n"
		"<directories> : all files under each directory became targets\n"
		"<directories> : named file becames a target\n"
		"");
				exit(0);
			case '?':
			default:
				break;
		}
	}

	argc -= optind;
	argv += optind;
	signal(SIGINT, restore_timestamps); /* ctrl-C */
	signal(SIGQUIT, restore_timestamps); /* ctrl-\ : ??? */
	signal(SIGTERM, restore_timestamps); /* kill.1 */

	if(init_array(&sys.domain, 512) == NULL)
	{
		char buf[512];
		sprintf(buf, "%s:realloc:out of memory", __func__);
		perror(buf);
		exit(-__LINE__);
	}

	if(filename == NULL)
	{
		for(c = 0; c < argc; c++)
			retreave(argv[c], NULL, NULL);
	}
	else
	{
		FILE *fd;
		char input[MAXPATHLEN + 1];
		int is_stdin = !strcmp("-", filename);

		if(is_stdin)
			fd = stdin;
		else if((fd = fopen(filename, "r")) == NULL)
			exit(-__LINE__);
		while(fgets(input, sizeof(input), fd))
		{ /* XXX: need to enhance tokenization... like -0 for xargs */
			char *p, **path = &p, *next = input;
			while((*path = strsep(&next, " \t\n")) != NULL)
				if(**path != '\0')
					retreave(*path, NULL, NULL);
		}
		if(!is_stdin)
			fclose(fd);
	}
	inoctl();
	return 0;
}
